//
// Copyright (c) 2008, John Sublett
// Licensed under the Academic Free License version 3.0
//
// History:
//   01 Jan 08  John Sublett  Creation
//

using sql

**
** DbModel is a collection of types and their relationships
** to be modeled in a relational database.
**
internal class DbModel
{
  **
  ** Add a type to the model.
  **
  Void addType(Type type)
  {
    types.add(type)
  }
  
  **
  ** Commit the model.  This must be called when all types have
  ** been added to the model.  A committed model may have types
  ** added or removed, but commit must be called again after
  ** the modifications are made.
  **
  Void commit()
  {
    // create dependency graph
    tg := TypeGraph.make
    types.each |Type t|
    {
      tg.addType(t)
      
      Type[] dependsOn := dependencies(t)
      dependsOn.each |Type dependency|
      {
        if (!types.contains(dependency))
          throw HavenErr.make("Missing dependency.  $t depends on "+dependency+".")
        tg.addDependency(t, dependency)
      }
    }
    
    // sort by dependency, this results in a list of types
    // in the order that the tables need to be created
    types = tg.sort
  }

  **
  ** List the types in the model.  The result is a readonly
  ** list of types in this model ordered by dependency.
  **
  Type[] listTypes()
  {
    return types.ro
  }

  **
  ** Get the list of Types that the specified
  ** dependent type depends on.
  **
  static Type[] dependencies(Type dependent)
  {
    depends := Type[,]
    dependent.fields.each |Field f|
    {
      if (f.of != sys::Uri.type) return
      targetType := f.facet("ref")
      if ((targetType != null) && !depends.contains(targetType))
        depends.add(targetType)
    }
    
    return depends
  }

  **
  ** Generate the DDL required to create the database schema
  ** for this model.
  **
  Str[] ddl(HavenService haven)
  {
    statements := Str[,]
    types.each |Type t, Int index|
    {
      statements.addAll(tableDef(haven, t).createTableDdl)
    }
    
    return statements
  }
  
  TableDef tableDef(HavenService haven, Type type)
  {
    TableDef def := tableDefs[type.qname]
    if (def == null)
    {
      def = TableDef.make(haven, type)
      tableDefs[type.qname] = def
    }
    
    return def
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  private Type[] types := Type[,]
  private Str:TableDef tableDefs := Str:TableDef[:]
}