//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   18 Aug 07  Brian Frank  Creation
//
package fanx.serial;

import fan.sys.*;
import fanx.util.*;

/**
 * ObjDecoder parses an object tree from an input stream.
 */
public class ObjDecoder
{

//////////////////////////////////////////////////////////////////////////
// Static
//////////////////////////////////////////////////////////////////////////

  public static Obj decode(String s)
  {
    return new ObjDecoder(InStream.makeForStr(s), null).readObj();
  }

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  /**
   * Construct for input stream.
   */
  public ObjDecoder(InStream in, Map options)
  {
    tokenizer = new Tokenizer(in);
    this.options = options;
    consume();
  }

//////////////////////////////////////////////////////////////////////////
// Parse
//////////////////////////////////////////////////////////////////////////

  /**
   * Read an object from the stream.
   */
  public final Obj readObj()
  {
    return readObj(null, null, true);
  }

  /**
   * obj := literal | simple | complex
   */
  private Obj readObj(Field curField, Type peekType, boolean root)
  {
    // literals are stand alone
    if (Token.isLiteral(curt))
    {
      Obj val = (Obj)tokenizer.val;
      consume();
      return val;
    }

    // [ is always list/map collection
    if (curt == Token.LBRACKET)
      return readCollection(curField, peekType);

    // at this point all remaining options must start
    // with a type signature - if peekType is non-null
    // then we've already read the type signature
    int line = tokenizer.line;
    Type t = (peekType != null) ? peekType : readType();

    // type:     type."type"
    // simple:   type(
    // list/map: type[
    // complex:  type || type{
    if (curt == Token.LPAREN)
      return readSimple(line, t);
    else if (curt == Token.DOT)
      return readTypeLiteral(line, t);
    else if (curt == Token.LBRACKET)
      return readCollection(curField, t);
    else
      return readComplex(line, t, root);
  }

  /**
   * typeLiteral := type "." "type"
   */
  private Obj readTypeLiteral(int line, Type t)
  {
    String errMsg = "Expected '.type' for type literal";
    consume(Token.DOT, errMsg );
    String id = consumeId(errMsg );
    if (!id.equals("type")) throw err(errMsg);
    return t;
  }

  /**
   * simple := type "(" str ")"
   */
  private Obj readSimple(int line, Type t)
  {
    // parse: type(str)
    consume(Token.LPAREN, "Expected ( in simple");
    Str str = consumeStr("Expected string literal for simple");
    consume(Token.RPAREN, "Expected ) in simple");

    // lookup the fromStr method
    t.finish();
    Method m = t.method("fromStr", false);
    if (m == null)
      throw err("Missing method: " + t.qname() + ".fromStr", line);

    // invoke parse method to translate into instance
    try
    {
      return m.invoke(null, new Object[] { str });
    }
    catch (ParseErr.Val e)
    {
      throw ParseErr.make(e.err().message() + " [Line " + line + "]").val;
    }
    catch (Throwable e)
    {
      throw ParseErr.make(e.toString() + " [Line " + line + "]", e).val;
    }
  }

  /**
   * complex := type [fields]
   * fields  := "{" field (eos field)* "}"
   * field   := name "=" obj
   */
  private Obj readComplex(int line, Type t, boolean root)
  {
    // make instance
    Obj obj = null;
    try
    {
      List args = null;
      if (root && options != null)
        args = (List)options.get(optMakeArgs);
      obj = (Obj)t.make(args);
    }
    catch (Throwable e)
    {
      throw IOErr.make("Cannot make " + t + ": " + e + " [Line " + line + "]", e).val;
    }

    // check for braces
    if (curt != Token.LBRACE) return obj;
    consume();

    // fields
    while (curt != Token.RBRACE)
    {
      // parse name=
      line = tokenizer.line;
      String name = consumeId("Expected field name identifier");
      consume(Token.EQ, "Expected '=' for field name/value pair");

      // resolve field
      Field field = t.field(name, false);
      if (field == null) throw err("Field not found: " + t.qname() + "." + name, line);

      // parse value
      Obj val = readObj(field, null, false);

      // set field value (skip const check)
      try
      {
        if (field.isConst().val)
          field.set(obj, OpUtil.toImmutable(val), false);
        else
          field.set(obj, val);
      }
      catch (Throwable e)
      {
        throw IOErr.make("Cannot set field " + t.qname() + "." + name + ": " + e + " [Line " + line + "]", e).val;
      }

      // eos
      endOfStmt(line);
    }
    consume(Token.RBRACE, "Expected '}'");

    return obj;
  }

//////////////////////////////////////////////////////////////////////////
// Collection
//////////////////////////////////////////////////////////////////////////

  /**
   * collection := list | map
   */
  private Obj readCollection(Field curField, Type t)
  {
    // opening [
    consume(Token.LBRACKET, "Expecting '['");

    // if this could be a map type signature:
    //    [qname:qname]
    //    [qname:qname][]
    //    [qname:qname][][] ...
    // or it could just be the type signature of
    // of a embedded simple, complex, or list
    Type peekType = null;
    if (curt == Token.ID && t == null)
    {
      // peek at the type
      peekType = readType();

      // if we have [mapType] then this is non-inferred type signature
      if (curt == Token.RBRACKET && peekType instanceof MapType)
      {
        t = peekType; peekType = null;
        consume();
        while (curt == Token.LRBRACKET) { consume(); t = t.toListOf(); }
        consume(Token.LBRACKET, "Expecting '['");
      }
    }

    // handle special case of [,]
    if (curt == Token.COMMA && peekType == null)
    {
      consume();
      consume(Token.RBRACKET, "Expecting ']'");
      return new List(toListOfType(t, curField, false));
    }

    // handle special case of [:]
    if (curt == Token.COLON && peekType == null)
    {
      consume();
      consume(Token.RBRACKET, "Expecting ']'");
      return new Map(toMapType(t, curField, false));
    }

    // read first list item or first map key
    Obj first = readObj(null, peekType, false);

    // now we can distinguish b/w list and map
    if (curt == Token.COLON)
      return readMap(toMapType(t, curField, true), first);
    else
      return readList(toListOfType(t, curField, true), first);
  }

  /**
   * list := "[" obj ("," obj)* "]"
   */
  private Obj readList(Type of, Obj first)
  {
    // setup accumulator
    Obj[] acc = new Obj[8];
    int n = 0;
    acc[n++] = first;

    // parse list items
    while (curt != Token.RBRACKET)
    {
      consume(Token.COMMA, "Expected ','");
      if (curt == Token.RBRACKET) break;
      if (n >= acc.length)
      {
        Obj[] temp = new Obj[n*2];
        System.arraycopy(acc, 0, temp, 0, n);
        acc = temp;
      }
      acc[n++] = readObj(null, null, false);
    }
    consume(Token.RBRACKET, "Expected ']'");

    // infer type if needed
    if (of == null) of = Type.common(acc, n);

    return new List(of, acc, n);
  }

  /**
   * map     := "[" mapPair ("," mapPair)* "]"
   * mapPair := obj ":" + obj
   */
  private Obj readMap(MapType mapType, Obj firstKey)
  {
    // setup accumulator
    Map.FanHashMap map = new Map.FanHashMap();

    // finish first pair
    consume(Token.COLON, "Expected ':'");
    map.put(firstKey, readObj(null, null, false));

    // parse map pairs
    while (curt != Token.RBRACKET)
    {
      consume(Token.COMMA, "Expected ','");
      if (curt == Token.RBRACKET) break;
      Obj key = readObj(null, null, false);
      consume(Token.COLON, "Expected ':'");
      Obj val = readObj(null, null, false);
      map.put(key, val);
    }
    consume(Token.RBRACKET, "Expected ']'");

    // infer type if necessary
    if (mapType == null)
    {
      int size = map.size();
      Type k = Type.common((Obj[])map.keySet().toArray(new Obj[size]), size);
      Type v = Type.common((Obj[])map.values().toArray(new Obj[size]), size);
      mapType = new MapType(k, v);
    }

    return new Map((MapType)mapType, map);
  }

  /**
   * Figure out the type of the list:
   *   1) if t was explicit then use it
   *   2) if we have field typed as a list, then use its definition
   *   3) if inferred is false, then drop back to list of Obj
   *   4) If inferred is true then return null and we'll infer the common type
   */
  private Type toListOfType(Type t, Field curField, boolean infer)
  {
    if (t != null) return t;
    if (curField != null)
    {
      Type ft = curField.of();
      if (ft instanceof ListType) return ((ListType)ft).v;
    }
    if (infer) return null;
    return Sys.ObjType;
  }

  /**
   * Figure out the map type:
   *   1) if t was explicit then use it (check that it was a map type)
   *   2) if we have field typed as a map , then use its definition
   *   3) if inferred is false, then drop back to Obj:Obj
   *   4) If inferred is true then return null and we'll infer the common key/val types
   */
  private MapType toMapType(Type t, Field curField, boolean infer)
  {
    if (t != null)
    {
      try { return (MapType)t; }
      catch (ClassCastException e) { throw err("Invalid map type: " + t); }
    }

    if (curField != null)
    {
      Type ft = curField.of();
      if (ft instanceof MapType) return (MapType)ft;
    }

    if (infer) return null;
    return defaultMapType;
  }
  private static final MapType defaultMapType = new MapType(Sys.ObjType, Sys.ObjType);

//////////////////////////////////////////////////////////////////////////
// Type
//////////////////////////////////////////////////////////////////////////

  /**
   * type    := listSig | mapSig1 | mapSig2 | qname
   * listSig := type "[]"
   * mapSig1 := type ":" type
   * mapSig2 := "[" type ":" type "]"
   *
   * Note: the mapSig2 with brackets is handled by the
   * method succinctly named readMapTypeOrCollection().
   */
  private Type readType() { return readType(false); }
  private Type readType(boolean lbracket)
  {
    Type t = readSimpleType();
    if (curt == Token.COLON)
    {
      consume();
      t = new MapType(t, readType());
    }
    while (curt == Token.LRBRACKET)
    {
      consume();
      t = t.toListOf();
    }
    return t;
  }

  /**
   * qname := podName "::" + typeName
   */
  private Type readSimpleType()
  {
    // parse qname
    int line = tokenizer.line;
    String podName = consumeId("Expected type signature");
    consume(Token.DOUBLE_COLON, "Expected :: in type qname");
    String typeName = consumeId("Expected type name");

    // resolve pod
    Pod pod = Pod.find(podName, false);
    if (pod == null) throw err("Pod not found: " + podName, line);

    // resolve type
    Type type = pod.findType(typeName, false);
    if (type == null) throw err("Type not found: " + podName + "::" + typeName, line);
    return type;
  }

//////////////////////////////////////////////////////////////////////////
// Error Handling
//////////////////////////////////////////////////////////////////////////

  /**
   * Create error reporting exception.
   */
  static RuntimeException err(String msg, int line)
  {
    return IOErr.make(msg + " [Line " + line + "]").val;
  }

  /**
   * Create exception based on tokenizers current line.
   */
  private RuntimeException err(String msg)
  {
    return err(msg, tokenizer.line);
  }

//////////////////////////////////////////////////////////////////////////
// Tokens
//////////////////////////////////////////////////////////////////////////

  /**
   * Consume the current token as a identifier.
   */
  private String consumeId(String expected)
  {
    verify(Token.ID, expected);
    String id = (String)tokenizer.val;
    consume();
    return id;
  }

  /**
   * Consume the current token as a Str literal.
   */
  private Str consumeStr(String expected)
  {
    verify(Token.STR_LITERAL, expected);
    Str id = (Str)tokenizer.val;
    consume();
    return id;
  }

  /**
   * Check that the current token matches the
   * specified type, and then consume it.
   */
  private void consume(int type, String expected)
  {
    verify(type, expected);
    consume();
  }

  /**
   * Check that the current token matches the specified
   * type, but do not consume it.
   */
  private void verify(int type, String expected)
  {
    if (curt != type)
      throw err(expected + ", not '" + Token.toString(curt) + "'");
  }

  /**
   * Consume the current token.
   */
  private void consume()
  {
    curt = tokenizer.next();
  }

  /**
   * Statements can be terminated with a semicolon, end of line or } end of block.
   */
  private void endOfStmt(int lastLine)
  {
    if (curt == Token.SEMICOLON) { consume(); return; }
    if (lastLine < tokenizer.line) return;
    if (curt == Token.RBRACE) return;
    throw err("Expected end of statement: semicolon, newline, or end of block; not '" + Token.toString(curt) + "'");
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  static final Str optMakeArgs = Str.make("makeArgs");

  Tokenizer tokenizer;    // tokenizer
  int curt;               // current token type
  Map options;            // decode option name/value pairs

}