//
// Copyright (c) 2008, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   25 Feb 08  Andy Frank  Split out from Buf
//

using Math = System.Math;
using System.IO;
using System.Security.Cryptography;
using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// MemBuf.
  /// </summary>
  public sealed class MemBuf : Buf
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public MemBuf(int capacity)
      : this(new byte[capacity], 0)
    {
    }

    public MemBuf(byte[] bytes)
      : this(bytes, bytes.Length)
    {
    }

    public MemBuf(byte[] bytes, int size)
    {
      this.m_buf  = bytes;
      this.m_pos  = 0;
      this.m_size = size;
      this.m_out  = new MemBufOutStream(this);
      this.m_in   = new MemBufInStream(this);
    }

  //////////////////////////////////////////////////////////////////////////
  // Obj
  //////////////////////////////////////////////////////////////////////////

    public override Type type() { return Sys.MemBufType; }

  //////////////////////////////////////////////////////////////////////////
  // Buf Support
  //////////////////////////////////////////////////////////////////////////

    internal override long getSize()
    {
      return m_size;
    }

    internal override void setSize(long x)
    {
      int newSize = (int)x;
      if (newSize > m_buf.Length)
      {
        byte[] temp = new byte[newSize];
        System.Array.Copy(m_buf, 0, temp, 0, m_buf.Length);
        m_buf = temp;
      }
      m_size = newSize;
    }

    internal override long getPos()
    {
      return m_pos;
    }

    internal override void setPos(long x)
    {
      this.m_pos = (int)x;
    }

    internal override int getByte(long pos)
    {
      return m_buf[(int)pos] & 0xFF;
    }

    internal override void setByte(long pos, int x)
    {
      m_buf[(int)pos] = (byte)x;
    }

    internal override void getBytes(long pos, byte[] dest, int off, int len)
    {
      System.Array.Copy(m_buf, pos, dest, off, len);
    }

    internal override void pipeTo(byte[] dst, int dstPos, int len)
    {
      if (m_pos + len > m_size) throw IOErr.make("Not enough bytes to write").val;
      System.Array.Copy(m_buf, m_pos, dst, dstPos, len);
      m_pos += len;
    }

    internal override void pipeTo(Stream dst, long lenLong)
    {
      int len = (int)lenLong;
      if (m_pos + len > m_size) throw IOErr.make("Not enough bytes to write").val;
      dst.Write(m_buf, m_pos, len);
      m_pos += len;
    }

    /*
    internal override void pipeTo(ByteBuffer dst, int len)
    {
      if (pos + len > size) throw IOErr.make("Not enough bytes to write").val;
      dst.put(buf, pos, len);
      pos += len;
    }
    */

    internal override void pipeFrom(byte[] src, int srcPos, int len)
    {
      grow(m_pos+len);
      System.Array.Copy(src, srcPos, m_buf, m_pos, len);
      m_pos += len;
      m_size = m_pos;
    }

    internal override long pipeFrom(Stream src, long lenLong)
    {
      int len = (int)lenLong;
      grow(m_pos+len);
      int read = src.Read(m_buf, m_pos, len);
      if (read == 0) return -1;
      m_pos  += read;
      m_size = m_pos;
      return read;
    }

    /*
    internal override int pipeFrom(ByteBuffer src, int len)
    {
      grow(pos+len);
      src.get(buf, pos, len);
      pos += len;
      size = pos;
      return len;
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Buf API
  //////////////////////////////////////////////////////////////////////////

    public override sealed Int capacity()
    {
      return Int.pos(m_buf.Length);
    }

    public override sealed void capacity(Int c)
    {
      int newCapacity = (int)c.val;
      if (newCapacity < m_size) throw ArgErr.make("capacity < size").val;
      byte[] temp = new byte[newCapacity];
      System.Array.Copy(m_buf, 0, temp, 0, newCapacity);
      m_buf = temp;
    }

    public override Buf trim()
    {
      if (m_size == m_buf.Length) return this;
      byte[] temp = new byte[m_size];
      System.Array.Copy(m_buf, 0, temp, 0, m_size);
      m_buf = temp;
      return this;
    }

    public override Str toHex()
    {
      char[] hexChars = Buf.hexChars;
      StringBuilder s = new StringBuilder(m_size*2);
      for (int i=0; i<m_size; ++i)
      {
        int b = m_buf[i] & 0xFF;
        s.Append(hexChars[b>>4]).Append(hexChars[b&0xf]);
      }
      return Str.make(s.ToString());
    }

    public override Str toBase64()
    {
      StringBuilder s = new StringBuilder(m_size*2);
      char[] base64chars = Buf.base64chars;
      int i = 0;

      // append full 24-bit chunks
      int end = m_size-2;
      for (; i<end; i += 3)
      {
        int c = m_buf[i];
        int n = (m_buf[i] << 16) + ((m_buf[i+1]) << 8) + m_buf[i+2];
        s.Append(base64chars[(n >> 18) & 0x3f]);
        s.Append(base64chars[(n >> 12) & 0x3f]);
        s.Append(base64chars[(n >> 6) & 0x3f]);
        s.Append(base64chars[n & 0x3f]);
      }

      // pad and encode remaining bits
      int rem = m_size - i;
      if (rem > 0)
      {
        int n = ((m_buf[i] & 0xff) << 10) | (rem == 2 ? ((m_buf[m_size-1] & 0xff) << 2) : 0);
        s.Append(base64chars[(n >> 12) & 0x3f]);
        s.Append(base64chars[(n >> 6) & 0x3f]);
        s.Append(rem == 2 ? base64chars[n & 0x3f] : '=');
        s.Append('=');
      }

      return Str.make(s.ToString());
    }

    public override Buf toDigest(Str algorithm)
    {
      string alg = algorithm.val;
      if (alg == "SHA-1") alg = "SHA1";  // to make .NET happy
      HashAlgorithm ha = HashAlgorithm.Create(alg);
      if (ha == null)
        throw ArgErr.make("Unknown digest algorthm: " + algorithm.val).val;
      return new MemBuf(ha.ComputeHash(m_buf, 0, m_size));
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    public byte[] bytes()
    {
      byte[] r = new byte[m_size];
      System.Array.Copy(m_buf, 0, r, 0, m_size);
      return r;
    }

    public void grow(int capacity)
    {
      if (m_buf.Length >= capacity) return;
      byte[] temp = new byte[Math.Max(capacity, m_size*2)];
      System.Array.Copy(m_buf, 0, temp, 0, m_size);
      m_buf = temp;
    }

  //////////////////////////////////////////////////////////////////////////
  // MemBufOutStream
  //////////////////////////////////////////////////////////////////////////

    internal class MemBufOutStream : OutStream
    {
      internal MemBufOutStream(MemBuf parent) { this.p = parent; }
      private MemBuf p;

      public override sealed OutStream write(Int v) { return w((int)v.val); }
      public override sealed OutStream w(int v)
      {
        if (p.m_pos+1 >= p.m_buf.Length) p.grow(p.m_pos+1);
        p.m_buf[p.m_pos++] = (byte)v;
        if (p.m_pos > p.m_size) p.m_size = p.m_pos;
        return this;
      }

      public override OutStream writeBuf(Buf other, Int n)
      {
        int len = (int)n.val;
        p.grow(p.m_pos+len);
        other.pipeTo(p.m_buf, p.m_pos, len);
        p.m_pos += len;
        if (p.m_pos > p.m_size) p.m_size = p.m_pos;
        return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // MemBufInStream
  //////////////////////////////////////////////////////////////////////////

    internal class MemBufInStream : InStream
    {
      internal MemBufInStream(MemBuf parent) { this.p = parent; }
      private MemBuf p;

      public override Int read() { int n = r(); return n < 0 ? null : Int.m_pos[n]; }
      public override int r()
      {
        if (p.m_pos >= p.m_size) return -1;
        return p.m_buf[p.m_pos++] & 0xFF;
      }

      public override Int readBuf(Buf other, Int n)
      {
        if (p.m_pos >= p.m_size) return null;
        int len = Math.Min(p.m_size-p.m_pos, (int)n.val);
        other.pipeFrom(p.m_buf, p.m_pos, len);
        p.m_pos += len;
        return Int.pos(len);
      }

      public override InStream unread(Int n) { return unread((int)n.val); }
      public override InStream unread(int n)
      {
        // unreading a buffer is a bit weird - the typical case
        // is that we are pushing back the byte we just read in
        // which case we can just rewind the position; however
        // if we pushing back a different byte then we need
        // to shift the entire buffer and insert the byte
        if (p.m_pos > 0 && p.m_buf[p.m_pos-1] == (byte)n)
        {
          p.m_pos--;
        }
        else
        {
          if (p.m_size+1 >= p.m_buf.Length) p.grow(p.m_size+1);
          System.Array.Copy(p.m_buf, p.m_pos, p.m_buf, p.m_pos+1, p.m_size);
          p.m_buf[p.m_pos] = (byte)n;
          p.m_size++;
        }
        return this;
      }

      public override Int peek()
      {
        if (p.m_pos >= p.m_size) return null;
        return Int.m_pos[p.m_buf[p.m_pos] & 0xFF];
      }

      public override Int skip(Int n)
      {
        int oldPos = p.m_pos;
        p.m_pos += (int)n.val;
        if (p.m_pos < p.m_size) return n;
        p.m_pos = p.m_size;
        return Int.pos(p.m_pos-oldPos);
      }

    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public byte[] m_buf;
    public int m_pos;
    public int m_size;

  }
}