//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Jan 06  Andy Frank  Creation
//

using System;
using System.IO;
using System.Reflection;
using Pod = Fan.Sys.Pod;
using Sys = Fan.Sys.Sys;
using SysProps = Fan.Sys.SysProps;
using Fanx.Emit;
using Fanx.Fcode;
using Fanx.Util;

namespace Fanx.Tools
{
  public abstract class Tool
  {

  //////////////////////////////////////////////////////////////////////////
  // Fan Assembly Loading
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Register the assembly resolver.
    /// </summary>
    protected Tool()
    {
      AppDomain domain = AppDomain.CurrentDomain;
      domain.AssemblyResolve += new ResolveEventHandler(resolveAssembly);
    }

    /// <summary>
    /// Get the assembly name from the specified fullName.  The fullName
    /// has the following format:
    ///
    /// <name>, Version=<major.minor.build.patch>, Culture=<...>, PublicKeyToken=<...>
    ///
    /// </summary>
    public static string getAssemblyName(string fullName)
    {
      int sep = fullName.IndexOf(',');
      if (sep == -1)
        return fullName;
      else
        return fullName.Substring(0, sep);
    }

    /// <summary>
    /// Resolve an assembly that cannot be found in
    /// the current AppDomain.
    /// </summary>
    public static Assembly resolveAssembly(object sender, ResolveEventArgs args)
    {
      string asmName = getAssemblyName(args.Name);
      AppDomain domain = AppDomain.CurrentDomain;

      // check if already loaded
      Assembly[] current = domain.GetAssemblies();
      for (int i=0; i<current.Length; i++)
        if (asmName == current[i].GetName().Name)
          return current[i];

      // otherwise load it from disk
      string libDir = FileUtil.combine(Sys.HomeDir, "lib", "net");
      string dll = FileUtil.combine(libDir, asmName + ".dll");

      FileInfo f = new FileInfo(dll);
      if (!f.Exists)
      {
        // check tmp dir
        string tmpDir = FileUtil.combine(Sys.HomeDir, "lib", "tmp");
        dll = FileUtil.combine(tmpDir, asmName + ".dll");
        f = new FileInfo(dll);
      }
      if (!f.Exists)
      {
        // not emitted yet, emit
        Pod pod = Pod.find(asmName, true);
        return FTypeEmit.emitPod(pod.fpod, true, null);
      }

      // the file may have been generated by another process, so
      // check if we need to emit to flush out things this process
      // will need from FTypeEmit
      if (!FTypeEmit.isEmitted(asmName))
      {
        Pod pod = Pod.find(asmName, true);
        return FTypeEmit.emitPod(pod.fpod, true, null);
      }

      BinaryReader fIn = new BinaryReader(f.OpenRead());
      byte[] asm = fIn.ReadBytes((int)f.Length);
      fIn.Close();
      if (asm.Length != f.Length)
        throw new Exception("Could not read " + dll + ": " + asm.Length + " != " + f.Length);

      Assembly result = domain.Load(asm);
      return result;
    }

    public static void sysInit(string fanHome)
    {
      if (isInit) return;

      SysProps.putProperty("fan.home", fanHome);

      // TODO - we need to cleanup the tmp dir on fan startup, this
      // seemed like the easiest spot to do it for now - eventually
      // we should be smarter and reuse the tmp dll if the src pod
      // modified time is < current
      try
      {
        string tmpDir = FileUtil.combine(Sys.HomeDir, "lib", "tmp");
        if (Directory.Exists(tmpDir))
        {
          string[] tmp = Directory.GetFiles(tmpDir);
          for (int i=0; i<tmp.Length; i++)
            File.Delete(tmp[i]);
        }
      }
      catch (Exception)
      {
        // In some cases this is actually ok - if you spawn a
        // new process and call fan.exe, the old process still
        // has a lock, though the dlls should be the same, so
        // it should be ok

        //System.Console.WriteLine(e);
      }
      isInit = true;
    }

    static bool isInit = false;

  }
}