//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   30 Jan 07  Brian Frank  Creation
//

**
** ThreadTest
**
class ThreadTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  Void testIdentity()
  {
    // make
    a := Thread.make("testSys.foo")
    b := Thread.make("testSys.bar")
    c := Thread.make
    d := Thread.make
    verifyErr(ArgErr.type) |,|  { Thread.make("testSys.foo") }
    verifyErr(ArgErr.type) |,|  { Thread.make(c.name) }
    verifyErr(NameErr.type) |,| { Thread.make("a b") }

    // name
    verifyEq(a.name, "testSys.foo")
    verifyEq(b.name, "testSys.bar")
    verify(c.name != d.name)

    // uri
// TODO
//    verifySame("/sys/threads/$a.name".toUri.resolve.obj, a)
//    verifySame("/sys/threads/$b.name".toUri.resolve.obj, b)

    // equals
    verify(a == a)
    verify(a != b)

    // hash
    verifyEq(a.hash, a.name.hash)
    verifyEq(b.hash, b.name.hash)

    // toStr
    verifyEq(a.toStr, "testSys.foo")
    verifyEq(d.name, d.toStr)

    // find
    verifySame(Thread.find("testSys.foo"), a)
    verifySame(Thread.find(d.name), d)
    verifyEq(Thread.find("testSys.forget it", false), null)
    verifyErr(UnknownThreadErr.type) |,| { Thread.find("testSys.forget it") }
    verifyErr(UnknownThreadErr.type) |,| { Thread.find("testSys.forget it", true) }

    // list
    verify(Thread.list.isRO)
    verifyEq(Thread.list.type, Thread[].type)
    verify(Thread.list.contains(a))
    verify(Thread.list.contains(b))
    verify(Thread.list.contains(c))
    verify(Thread.list.contains(d))

    // current
    verifyEq(Thread.current.name, "main")
    verifyEq(Thread.current.isNew, false)
    verifyEq(Thread.current.isRunning, true)
    verifyEq(Thread.current.isDead, false)
  }

//////////////////////////////////////////////////////////////////////////
// Count
//////////////////////////////////////////////////////////////////////////

  Void testCount()
  {
    // create it
    t := ThreadCount.make
    verifyEq(t.name, "ThreadCount")
    verifyEq(t.isNew, true)
    verifyEq(t.isRunning, false)
    verifyEq(t.isDead, false)
    verifySame(Thread.find("ThreadCount"), t)
    verifyErr(ArgErr.type) |,| { ThreadCount.make }

    // stop it
    t.stop
    verifyEq(t.isNew, false)
    verifyEq(t.isRunning, false)
    verifyEq(t.isDead, true)
    verifyEq(Thread.find("ThreadCount", false), null)
    old := t

    // create new one
    t = ThreadCount.make
    verify(t !== old)
    verifyEq(t.name, "ThreadCount")
    verifyEq(t.isNew, true)

    // start it
    t.start
    verifyEq(t.isNew, false)
    verifyEq(t.isRunning, true)
    verifyEq(t.isDead, false)
    verifySame(Thread.find("ThreadCount"), t)
    verifyErr(Err.type) |,| { t.start }

    // join until counter done
    Thread.sleep(500ms)
    verifyEq(t.isNew, false)
    verifyEq(t.isRunning, false)
    verifyEq(t.isDead, true)
    verifyEq(Thread.find("ThreadCount", false), null)
    verifyErr(Err.type) |,| { t.start }
    t.join  // ignored
    t.stop  // ignored
  }

//////////////////////////////////////////////////////////////////////////
// Messaging
//////////////////////////////////////////////////////////////////////////

  Void testMessaging()
  {
    // verify run is checked for immutable
    verifyErr(NotImmutableErr.type) |,|
    {
      Thread.make("testSys.testMessaging") |Thread thread| { fail }
    }

    // create thread which takes an Int msg and doubles it
    t := Thread.make("testSys.testMessaging", &runMessaging)
    t.start

    // verify this thread can't call t's loop
    verifyErr(Err.type) |,| { t.loop |Obj m->Obj| { return m } }

    // send some sync messages
    10.times |Int i|
    {
      Int r := t.sendSync(i)
      verifyEq(r, i*2)
    }

    // send serialized req/res
    Msg req := Msg { a="a"; b="b" }
    Msg res := t.sendSync(req)
    verifyEq(req, Msg { a="a"; b="b" })
    verifyEq(res, Msg { a="(a)"; b="(b)"; c="(null)" })

    // verify exception raised to me
    verifyErr(IOErr.type) |,| { t.sendSync("throw") }

    // verify messages must immutable
    verifyErr(IOErr.type) |,| { t.sendSync(this) }
    verifyErr(IOErr.type) |,| { t.sendAsync(this) }
    verifyErr(IOErr.type) |,| { t.sendSync("mutable") }
    //t.sendAsync("mutable")
    //t.sendAsync("throw")
    t.sendSync("die")

    verify(t.isDead)
  }

  static Void runMessaging(Thread t)
  {
    t.loop |Obj msg->Obj|
    {
      if (msg is Int) return (Int)msg * 2

      if (msg is Msg)
      {
        res := msg as Msg
        res.a = "($res.a)"
        res.b = "($res.b)"
        res.c = "($res.c)"
        return res
      }

      switch (msg)
      {
        case "throw":    throw IOErr.make()
        case "mutable":  return ThreadTest.make
        case "die":      t.stop; return null
      }
      return null
    }
  }

//////////////////////////////////////////////////////////////////////////
// Join
//////////////////////////////////////////////////////////////////////////

  Void testJoin()
  {
    t := Thread.make(null, &runJoin("six")).start
    verifyEq(t.join(2sec), 6)
    verifyEq(t.join(2sec), null)

    t = Thread.make(null, &runJoin("mutable")).start
    verifyEq(t.join(2sec)->readAllStr, "abc")
    verifyEq(t.join(2sec), null)
  }

  static Obj runJoin(Str mode)
  {
    switch (mode)
    {
      case "six":     return 6
      case "mutable": return Buf.make.print("abc").flip
      default:        throw UnsupportedErr.make
    }
  }

//////////////////////////////////////////////////////////////////////////
// Sync Stop
//////////////////////////////////////////////////////////////////////////

  Void testSyncStop()
  {
    verifySyncStop(true)
    verifySyncStop(false)
  }

  Void verifySyncStop(Bool explicitStop)
  {
    x := Thread.make(null, &runSyncStopMain).start
    a := Thread.make(null, &runSyncStop(x)).start
    b := Thread.make(null, &runSyncStop(x)).start
    c := Thread.make(null, &runSyncStop(x)).start

    if (explicitStop)
    {
      Thread.sleep(50ms)
      x.stop
    }
    else
    {
      x.join
    }

    t1 := Duration.now
    verify(a.join is InterruptedErr)
    verify(b.join is InterruptedErr)
    verify(c.join is InterruptedErr)
    t2 := Duration.now
    verify(t2-t1 < 50ms)
  }

  static Void runSyncStopMain(Thread t)
  {
    try
    {
      Thread.sleep(100ms)
    }
    catch (InterruptedErr e)
    {
    }
  }

  static Obj runSyncStop(Thread main, Thread t)
  {
    try
    {
      main.sendSync("x")
      return null
    }
    catch (Err e)
    {
      return e
    }
  }

//////////////////////////////////////////////////////////////////////////
// OnCallbacks
//////////////////////////////////////////////////////////////////////////

  Void testOnCallbacks()
  {
    Sys.ns.create(`/testSys/onStart`, false)
    Sys.ns.create(`/testSys/run`, false)
    Sys.ns.create(`/testSys/onStop`, false)
    try
    {
      3.times |Int mode|
      {
        t := ThreadCallbacks.make(mode)

        t.start
        Thread.sleep(20ms)
        verifyEq(Sys.ns[`/testSys/onStart`], true)

        t.stop
        Thread.sleep(20ms)
        verifyEq(Sys.ns[`/testSys/onStart`], true)
        verifyEq(Sys.ns[`/testSys/run`],     mode >= 1)
        verifyEq(Sys.ns[`/testSys/onStop`],  true)
      }
    }
    finally
    {
      Sys.ns.delete(`/testSys/onStart`)
      Sys.ns.delete(`/testSys/run`)
      Sys.ns.delete(`/testSys/onStop`)
    }
  }

//////////////////////////////////////////////////////////////////////////
// Timer
//////////////////////////////////////////////////////////////////////////

  Void testTimer()
  {
    try
    {
      x := Thread.make(null, &runTimerTest)
      x.start.join
      Thread.sleep(20ms)
      Obj[][] results := Sys.ns[`/testSys/timer`]
      verifyTimer(results[0], 100ms, "100ms")
      verifyTimer(results[1], 100ms, "100ms repeat")
      verifyTimer(results[2], 200ms, "200ms")
      verifyTimer(results[3], 200ms, "100ms repeat")
      verifyTimer(results[4], 300ms, "300ms")
      verifyTimer(results[5], 400ms, "400ms")
      verifyTimer(results[6], 500ms, "500ms")
    }
    finally
    {
      Sys.ns.delete(`/testSys/timer`)
    }
  }

  Void verifyTimer(Obj[] results, Duration d, Str msg)
  {
    Duration a := results[0]
    verify(a.ticks.minus(d.ticks).abs < 20ms.ticks)
    verifyEq(results[1], msg)
  }

  static Obj runTimerTest(Thread t)
  {
    t.sendLater(100ms, "100ms")
    t.sendLater(200ms, "200ms")
    t.sendLater(300ms, "300ms")
    t.sendLater(400ms, "400ms")
    ticket := t.sendLater(100ms, "100ms repeat", true)
    t.sendLater(500ms, "500ms")

    start := Duration.now
    results := [,]
    t.loop |Obj msg|
    {
      results.add([Duration.now-start, msg])
      if (msg == "300ms") t.cancelLater(ticket)
      if (msg == "500ms") t.stop
    }

    Sys.ns.create(`/testSys/timer`, results)
    return null
  }

//////////////////////////////////////////////////////////////////////////
// Thread Locals
//////////////////////////////////////////////////////////////////////////

  Void testLocals()
  {
    Thread.locals["testSys.x"] = "main"
    verifyEq(Thread.locals["testSys.x"], "main")

    a := Thread.make(null, &runLocal("alpha")).start
    b := Thread.make(null, &runLocal("beta")).start
    c := Thread.make(null, &runLocal("gamma")).start

    verifyEq(a.join, "alpha")
    verifyEq(b.join, "beta")
    verifyEq(c.join, "gamma")
    verifyEq(Thread.locals["testSys.x"], "main")
  }

  static Obj runLocal(Str val)
  {
    r := val
    3.times |,|
    {
      Thread.locals["testSys.x"] = val
      if (Thread.locals["testSys.x"] != val)
      {
        Err.make.trace
        r = "bad bad bad"
      }
      Thread.sleep(10ms)
    }
    return r
  }

//////////////////////////////////////////////////////////////////////////
// Trace
//////////////////////////////////////////////////////////////////////////

  Void testTrace()
  {
    buf := Buf.make

    // trace new and dead thread
    t := Thread.make(null)
    t.trace(buf.out)
    t.stop
    t.trace(buf.out)
    verifyEq(buf.size, 0)

    // trace current
    traceA(buf.out)

    lines := buf.flip.readAllLines
    start := lines.find |Str s->Bool| { return s.contains("traceB") }
    verify(start != null)
    i := lines.index(start)
    verify(lines[i+0].contains("testSys::ThreadTest.traceB"))
    verify(lines[i+1].contains("testSys::ThreadTest.traceA"))
    verify(lines[i+2].contains("testSys::ThreadTest.testTrace"))
  }

  Void traceA(OutStream out) { traceB(out) }
  Void traceB(OutStream out) { Thread.current.trace(out) }

}

**************************************************************************
** ThreadCount
**************************************************************************

const class ThreadCount : Thread
{
  new make() : super("ThreadCount") {}

  override Obj run()
  {
    3.times |Int c| { sleep(100ms) }
    return null
  }
}

**************************************************************************
** Msg
**************************************************************************

@serializable
class Msg
{
  override Int hash() { return toStr.hash }
  override Bool equals(Obj obj) { return obj is Msg && obj.toStr == toStr }
  override Str toStr() { return "a=$a b=$b c=$c" }
  Str a
  Str b
  Str c
}

**************************************************************************
** ThreadCallbacks
**************************************************************************

const class ThreadCallbacks : Thread
{
  new make(Int mode) : super(null) { this.mode = mode }

  override Obj run()
  {
    Sys.ns.put(`/testSys/run`, true)
    if (mode == 1) ThreadTest.make.fail
    loop |Obj req->Obj|
    {
      echo("  req: $req")
      return null
    }
    return null
  }

  override Void onStart()
  {
    Sys.ns.put(`/testSys/onStart`, true)
    if (mode == 0) ThreadTest.make.fail
  }

  override Void onStop()
  {
    Sys.ns.put(`/testSys/onStop`, true)
    if (mode == 2) ThreadTest.make.fail
  }

  const Int mode
}
